# SPDX-FileCopyrightText: Copyright (c) 2025 Cisco and/or its affiliates.
# SPDX-License-Identifier: Apache-2.0

import argparse
import asyncio
import time

import agp_bindings
from agp_bindings import GatewayConfig


class color:
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


def format_message(message1, message2):
    return f"{color.BOLD}{color.CYAN}{message1.capitalize()}{color.END}\t {message2}"


async def run_client(local_id, remote_id, message, address, iterations, enable_opentelemetry: bool):
    # init tracing
    agp_bindings.init_tracing(log_level="debug", enable_opentelemetry=enable_opentelemetry)

    # Split the IDs into their respective components
    try:
        local_organization, local_namespace, local_agent = local_id.split("/")
    except ValueError:
        print("Error: IDs must be in the format organization/namespace/agent.")
        return

    # Define the service based on the local agent
    gateway = agp_bindings.Gateway()

    # Configure gateway
    config = GatewayConfig(endpoint=address, insecure=True)
    gateway.configure(config)

    # Connect to the gateway server
    local_agent_id = await gateway.create_agent(
        local_organization, local_namespace, local_agent
    )

    # Connect to the service and subscribe for the local name
    _ = await gateway.connect()
    await gateway.subscribe(
        local_organization, local_namespace, local_agent, local_agent_id
    )

    if message:
        if not iterations:
            iterations = 1

        # Split the IDs into their respective components
        try:
            remote_organization, remote_namespace, remote_agent = remote_id.split("/")
        except ValueError:
            print("Error: IDs must be in the format organization/namespace/agent.")
            return

        # Create a route to the remote ID
        await gateway.set_route(remote_organization, remote_namespace, remote_agent)

        for i in range (0, iterations):
            try:
                # Send the message
                await gateway.publish(
                    message.encode(), remote_organization, remote_namespace, remote_agent
                )
                print(format_message(f"{local_agent.capitalize()} sent:", message))

                # Wait for a reply
                src, msg = await gateway.receive()
                print(format_message(f"{local_agent.capitalize()} received:", msg.decode()))
            except Exception as e:
                print("received error: ", e)

            time.sleep(5)
    else:
        # Wait for a message and reply in a loop
        while True:
            src, msg = await gateway.receive()
            print(format_message(f"{local_agent.capitalize()} received:", msg.decode()))

            ret = f"Echo from {local_agent}: {msg.decode()}"

            await gateway.publish_to(ret.encode(), src)
            print(format_message(f"{local_agent.capitalize()} replies:", ret))


def main():
    parser = argparse.ArgumentParser(
        description="Command line client for message passing."
    )
    parser.add_argument(
        "-l",
        "--local",
        type=str,
        help="Local ID in the format organization/namespace/agent.",
    )
    parser.add_argument(
        "-r",
        "--remote",
        type=str,
        help="Remote ID in the format organization/namespace/agent.",
    )
    parser.add_argument("-m", "--message", type=str, help="Message to send.")
    parser.add_argument(
        "-g",
        "--gateway",
        type=str,
        help="Gateway address.",
        default="http://127.0.0.1:46357",
    )
    parser.add_argument("-i", "--iterations", type=int, help="Number of messages to send, one per second.")
    parser.add_argument(
        "-t",
        "--enable-opentelemetry",
        action="store_true",
        default=False,
        help="Enable OpenTelemetry tracing.",
    )

    args = parser.parse_args()

    # Run the client with the specified local ID, remote ID, and optional message
    asyncio.run(run_client(args.local, args.remote, args.message, args.gateway, args.iterations, args.enable_opentelemetry))


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright (c) 2025 Cisco and/or its affiliates.
# SPDX-License-Identifier: Apache-2.0

# import the contents of the Rust library into the Python extension
from typing import Optional, Tuple
from ._agp_bindings import (
    PyGatewayConfig as GatewayConfig,
    PyService,
    PyAgentClass,
    create_agent,
    connect,
    disconnect,
    publish,
    receive,
    init_tracing,
    subscribe,
    unsubscribe,
    serve,
    set_route,
    remove_route,
)
from ._agp_bindings import __all__

# optional: include the documentation from the Rust module
from ._agp_bindings import __doc__  # noqa: F401

class Gateway:
    def __init__(self, name="gateway/agent"):
        """
        Create a new Gateway instance.

        Args:
            name (str): The name of the Gateway service. Default is "gateway/agent".

        Returns:
            Gateway: A new Gateway instance
        """
        self.svc = PyService(name)

    def configure(self, config):
        """
        Configure the gateway.

        Args:
            config (GatewayConfig): The gateway configuration class.

        Returns:
            None
        """

        self.svc.configure(config)

    async def create_agent(
        self, organization, namespace, agent, id: Optional[int] = None
    ) -> int:
        """
        Create a new agent.

        Args:
            organization (str): The organization of the agent.
            namespace (str): The namespace of the agent.
            agent (str): The name of the agent.

        Returns:
            None
        """

        return await create_agent(self.svc, organization, namespace, agent, id)

    async def serve(self):
        """
        Serve the Gateway service.

        Args:
            None

        Returns:
            None
        """

        await serve(self.svc)

    async def connect(self) -> int:
        """
        Connect to a remote gateway service.

        Args:
            None

        Returns:
            int: The connection ID.
        """

        self.conn_id = await connect(self.svc)

        return self.conn_id
    
    async def disconnect(self):
        """
        disconnect from a remote gateway service.

        Args:
            None

        Returns:
            None

        """

        await disconnect(self.svc, self.conn_id)

    async def set_route(self, organization, namespace, agent, id: Optional[int] = None):
        """
        Set route for outgoing messages via the connected gateway.

        Args:
            organization (str): The organization of the agent.
            namespace (str): The namespace of the agent.
            agent (str): The name of the agent.

        Returns:
            None
        """

        name = PyAgentClass(organization, namespace, agent)
        await set_route(self.svc, self.conn_id, name, id)

    async def remove_route(self, organization, namespace, agent, id: Optional[int] = None):
        """
        Remove route for outgoing messages via the connected gateway.

        Args:
            organization (str): The organization of the agent.
            namespace (str): The namespace of the agent.
            agent (str): The name of the agent.

        Returns:
            None
        """

        name = PyAgentClass(organization, namespace, agent)
        await remove_route(self.svc, self.conn_id, name, id)

    async def subscribe(self, organization, namespace, agent, id=None):
        """
        Subscribe to receive messages for the given agent.

        Args:
            organization (str): The organization of the agent.
            namespace (str): The namespace of the agent.
            agent (str): The name of the agent.
            id (int): The ID of the agent.

        Returns:
            None
        """

        sub = PyAgentClass(organization, namespace, agent)
        await subscribe(self.svc, self.conn_id, sub, id)

    async def unsubscribe(self, organization, namespace, agent, id=None):
        """
        Unsubscribe from receiving messages for the given agent.

        Args:
            organization (str): The organization of the agent.
            namespace (str): The namespace of the agent.
            agent (str): The name of the agent.
            id (int): The ID of the agent.

        Returns:
            None
        """
        unsub = PyAgentClass(organization, namespace, agent)
        await unsubscribe(self.svc, self.conn_id, unsub, id)

    async def publish(self, msg, organization, namespace, agent):
        """
        Publish a message to an agent via normal matching in subscription table.

        Args:
            msg (str): The message to publish.
            organization (str): The organization of the agent.
            namespace (str): The namespace of the agent.
            agent (str): The name of the agent.

        Returns:
            None
        """

        dest = PyAgentClass(organization, namespace, agent)
        await publish(self.svc, 1, msg, dest, None)

    async def publish_to(self, msg, agent):
        """
        Publish a message to an agent via the connected gateway.

        Args:
            msg (str): The message to publish.
            agent (Agent): The agent to publish to.

        Returns:
            None
        """

        await publish(self.svc, 1, msg, agent=agent)

    async def receive(self) -> Tuple[any, bytes]:
        """
        Receive a message from the connected gateway.

        Returns:
            tuple: The source agent and the message.

        """

        source, msg = await receive(self.svc)
        return source, msg



# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins
import typing

class PyAgentClass:
    r"""
    agent class
    """
    def __new__(cls,agent_org:builtins.str, agent_ns:builtins.str, agent_class:builtins.str): ...
    ...

class PyAgentSource:
    r"""
    packet source with encoded agent information
    plus incoming connection
    """
    def __new__(cls,org:builtins.int, ns:builtins.int, class:builtins.int, id:builtins.int, connection:builtins.int): ...
    ...

class PyGatewayConfig:
    r"""
    gatewayconfig class
    """
    def __new__(cls,endpoint:builtins.str, insecure:builtins.bool=False, insecure_skip_verify:builtins.bool=False, tls_ca_path:typing.Optional[builtins.str]=None, tls_ca_pem:typing.Optional[builtins.str]=None, tls_cert_path:typing.Optional[builtins.str]=None, tls_key_path:typing.Optional[builtins.str]=None, tls_cert_pem:typing.Optional[builtins.str]=None, tls_key_pem:typing.Optional[builtins.str]=None, basic_auth_username:typing.Optional[builtins.str]=None, basic_auth_password:typing.Optional[builtins.str]=None): ...
    ...

class PyService:
    def __new__(cls,id:builtins.str): ...
    def configure(self, config:PyGatewayConfig) -> None:
        ...


def connect(svc:PyService) -> typing.Any:
    ...

def create_agent(svc:PyService, agent_org:builtins.str, agent_ns:builtins.str, agent_class:builtins.str, agent_id:typing.Optional[builtins.int]=None) -> typing.Any:
    ...

def disconnect(svc:PyService, conn:builtins.int) -> typing.Any:
    ...

def publish(svc:PyService, fanout:builtins.int, blob:typing.Sequence[builtins.int], name:typing.Optional[PyAgentClass]=None, id:typing.Optional[builtins.int]=None, agent:typing.Optional[PyAgentSource]=None) -> typing.Any:
    ...

def receive(svc:PyService) -> typing.Any:
    ...

def remove_route(svc:PyService, conn:builtins.int, name:PyAgentClass, id:typing.Optional[builtins.int]=None) -> typing.Any:
    ...

def serve(svc:PyService) -> typing.Any:
    ...

def set_route(svc:PyService, conn:builtins.int, name:PyAgentClass, id:typing.Optional[builtins.int]=None) -> typing.Any:
    ...

def subscribe(svc:PyService, conn:builtins.int, name:PyAgentClass, id:typing.Optional[builtins.int]=None) -> typing.Any:
    ...

def unsubscribe(svc:PyService, conn:builtins.int, name:PyAgentClass, id:typing.Optional[builtins.int]=None) -> typing.Any:
    ...




# SPDX-FileCopyrightText: Copyright (c) 2025 Cisco and/or its affiliates.
# SPDX-License-Identifier: Apache-2.0

import argparse
import asyncio
from signal import SIGINT

import agp_bindings
from agp_bindings import GatewayConfig

# Create a service
gateway = agp_bindings.Gateway()


async def run_server(address: str, enable_opentelemetry: bool):
    # init tracing
    agp_bindings.init_tracing(log_level="debug", enable_opentelemetry=enable_opentelemetry)

    # Configure gateway
    config = GatewayConfig(endpoint=address, insecure=True)
    gateway.configure(config)

    # Run as server
    await gateway.serve()


async def main():
    parser = argparse.ArgumentParser(
        description="Command line client for gateway server."
    )
    parser.add_argument(
        "-g", "--gateway", type=str, help="Gateway address.", default="127.0.0.1:12345"
    )
    parser.add_argument(
        "--enable-opentelemetry",
        "-t",
        action="store_true",
        default=False,
        help="Enable OpenTelemetry tracing.",
    )

    args = parser.parse_args()

    # Create an asyncio event to keep the loop running until interrupted
    stop_event = asyncio.Event()

    # Define a shutdown handler to set the event when interrupted
    def shutdown():
        print("\nShutting down...")
        stop_event.set()

    # Register the shutdown handler for SIGINT
    loop = asyncio.get_running_loop()
    loop.add_signal_handler(SIGINT, shutdown)

    # Run the client task
    client_task = asyncio.create_task(run_server(args.gateway, args.enable_opentelemetry))

    # Wait until the stop event is set
    await stop_event.wait()

    # Cancel the client task
    client_task.cancel()
    try:
        await client_task
    except asyncio.CancelledError:
        pass


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        print("Program terminated by user.")



# SPDX-FileCopyrightText: Copyright (c) 2025 Cisco and/or its affiliates.
# SPDX-License-Identifier: Apache-2.0

import asyncio
import time
import pytest
import pytest_asyncio
import agp_bindings

# create svcs
svc_server = agp_bindings.PyService("gateway/server")
svc_server.configure(agp_bindings.GatewayConfig(
    endpoint="0.0.0.0:12345", 
    insecure=True
))


@pytest_asyncio.fixture(scope="module")
async def server():
    # init tracing
    agp_bindings.init_tracing()

    # run gateway server in background
    await agp_bindings.serve(svc_server)

    # wait for the server to start
    await asyncio.sleep(1)

@pytest.mark.asyncio
async def test_end_to_end(server):
    # create 2 clients, Alice and Bob
    svc_alice = agp_bindings.PyService("gateway/alice")
    svc_alice.configure(agp_bindings.GatewayConfig(
        endpoint="http://127.0.0.1:12345"
    ))

    svc_bob = agp_bindings.PyService("gateway/bob")
    svc_bob.configure(agp_bindings.GatewayConfig(
        endpoint="http://127.0.0.1:12345"
    ))

    # connect to the gateway server
    await agp_bindings.create_agent(svc_alice, "cisco", "default", "alice", 1234)
    await agp_bindings.create_agent(svc_bob, "cisco", "default", "bob", 1234)

    # connect to the service
    conn_id_alice = await agp_bindings.connect(svc_alice)
    conn_id_bob = await agp_bindings.connect(svc_bob)

    # subscribe alice and bob
    alice_class = agp_bindings.PyAgentClass("cisco", "default", "alice")
    bob_class = agp_bindings.PyAgentClass("cisco", "default", "bob")
    await agp_bindings.subscribe(svc_alice, conn_id_alice, alice_class, 1234)
    await agp_bindings.subscribe(svc_bob, conn_id_bob, bob_class, 1234)

    # set routes
    await agp_bindings.set_route(svc_alice, conn_id_alice, bob_class, None)

    # wait for the routes to be set
    # TODO remove this sleep
    await asyncio.sleep(1)

    # send msg from Alice to Bob
    msg = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    dest = agp_bindings.PyAgentClass("cisco", "default", "bob")
    await agp_bindings.publish(svc_alice, 1, msg, dest, None)

    # receive message from Alice
    source, msg_rcv = await agp_bindings.receive(svc_bob)

    # check if the message is correct
    assert msg_rcv == bytes(msg)

    # reply to Alice
    await agp_bindings.publish(svc_bob, 1, msg_rcv, agent=source)

    # wait for message
    source, msg_rcv = await agp_bindings.receive(svc_alice)

    print(msg_rcv)

    # check if the message is correct
    assert msg_rcv == bytes(msg)

    # disconnect alice
    await agp_bindings.disconnect(svc_alice, conn_id_alice)

    # disconnect bob
    await agp_bindings.disconnect(svc_bob, conn_id_bob)

@pytest.mark.asyncio
async def test_gateway_wrapper(server):
    # create new gateway object
    gateway1 = agp_bindings.Gateway("gateway/gateway1")
    gateway1.configure(agp_bindings.GatewayConfig(
        endpoint="http://127.0.0.1:12345", 
        insecure=True
    ))

    org = "cisco"
    ns = "default"
    agent1 = "gateway1"

    # create local agent 1
    await gateway1.create_agent(org, ns, agent1)

    # Connect to the gateway server
    local_agent_id1 = await gateway1.create_agent(org, ns, agent1)

    # Connect to the service and subscribe for the local name
    _ = await gateway1.connect()

    # disconnect and reconnect
    await gateway1.disconnect()
    time.sleep(1)

    _ = await gateway1.connect()

    await gateway1.subscribe(org, ns, agent1, local_agent_id1)

    # create second local agent
    gateway2 = agp_bindings.Gateway("gateway/gateway2")
    gateway2.configure(agp_bindings.GatewayConfig(
        endpoint="http://127.0.0.1:12345", 
        insecure=True
    ))

    agent2 = "gateway2"

    local_agent_id2 = await gateway2.create_agent(org, ns, agent2)

    # Connect to gateway server
    _ = await gateway2.connect()
    await gateway2.subscribe(org, ns, agent2, local_agent_id2)

    # set route
    await gateway2.set_route("cisco", "default", agent1)

    # wait for the routes to be set
    # TODO remove this sleep
    await asyncio.sleep(1)

    # publish message
    msg = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    await gateway2.publish(msg, org, ns, agent1)

    # receive message
    source, msg_rcv = await gateway1.receive()

    # check if the message is correct
    assert msg_rcv == bytes(msg)

    # reply to Alice
    await gateway1.publish_to(msg_rcv, source)

    # wait for message
    source, msg_rcv = await gateway2.receive()

    # check if the message is correct
    assert msg_rcv == bytes(msg)

